@@ -136,6 +136,7 @@ struct ModuleState {
     tables: Vec<TableType>,
     memories: Vec<MemoryType>,
     globals: Vec<GlobalType>,
+    num_imported_globals: u32,
     element_types: Vec<Type>,
     data_count: Option<u32>,
     code_type_indexes: Vec<u32>, // pointer into `types` above
@@ -257,7 +258,8 @@ impl TypeDef {
 }
 
 struct ModuleType {
-    type_size: u32,
+    imports_size: u32,
+    exports_size: u32,
     imports: HashMap<String, EntityType>,
     exports: HashMap<String, EntityType>,
 }
@@ -598,12 +600,10 @@ impl Validator {
                     let ty = self.import_entry_type(&e.ty)?;
                     exports.push(self.offset, e.name, None, ty, &mut self.types, "export")?;
                 }
+                combine_type_sizes(self.offset, imports.type_size, exports.type_size)?;
                 TypeDef::Module(ModuleType {
-                    type_size: combine_type_sizes(
-                        self.offset,
-                        imports.type_size,
-                        exports.type_size,
-                    )?,
+                    imports_size: imports.type_size,
+                    exports_size: exports.type_size,
                     imports: imports.set,
                     exports: exports.set,
                 })
@@ -833,6 +833,7 @@ impl Validator {
             }
             ImportSectionEntryType::Global(ty) => {
                 state.globals.push(ty);
+                state.num_imported_globals += 1;
                 (state.globals.len(), MAX_WASM_GLOBALS, "globals")
             }
             ImportSectionEntryType::Instance(type_idx) => {
@@ -924,7 +925,9 @@ impl Validator {
                     }
                     (EntityType::Global(ty), ExternalKind::Global) => {
                         let ty = ty.clone();
-                        self.cur.state.assert_mut().globals.push(ty);
+                        let state = self.cur.state.assert_mut();
+                        state.num_imported_globals += 1;
+                        state.globals.push(ty);
                     }
                     (EntityType::Instance(ty), ExternalKind::Instance) => {
                         let ty = *ty;
@@ -1018,7 +1021,7 @@ impl Validator {
         // accounting for the imports on the original module, but that should be
         // ok for now since it's only used to limit the size of larger types.
         let instance_ty = InstanceType {
-            type_size: ty.type_size,
+            type_size: ty.exports_size,
             exports: ty.exports.clone(),
         };
         self.cur.state.assert_mut().instances.push(self.types.len());
@@ -1265,6 +1268,11 @@ impl Validator {
             Operator::V128Const { .. } => Type::V128,
             Operator::GlobalGet { global_index } => {
                 let global = self.get_global(global_index)?;
+                if global_index >= self.cur.state.num_imported_globals {
+                    return self.create_error(
+                        "constant expression required: global.get of locally defined global",
+                    );
+                }
                 if global.mutable {
                     return self.create_error(
                         "constant expression required: global.get of mutable global",
@@ -1563,7 +1571,7 @@ impl Validator {
         // size. This is primarily here for the module linking proposal, and
         // we'll record this in the module type below if we're part of a nested
         // module.
-        let type_size = combine_type_sizes(
+        combine_type_sizes(
             self.offset,
             self.cur.state.imports.type_size,
             self.cur.state.exports.type_size,
@@ -1576,7 +1584,8 @@ impl Validator {
         if let Some(mut parent) = self.parents.pop() {
             let module_type = self.types.len();
             self.types.push(TypeDef::Module(ModuleType {
-                type_size,
+                imports_size: self.cur.state.imports.type_size,
+                exports_size: self.cur.state.exports.type_size,
                 imports: self.cur.state.imports.set.clone(),
                 exports: self.cur.state.exports.set.clone(),
             }));
@@ -1587,6 +1596,50 @@ impl Validator {
     }
 }
 
+impl WasmModuleResources for Validator {
+    type FuncType = crate::FuncType;
+
+    fn table_at(&self, at: u32) -> Option<TableType> {
+        self.cur.state.table_at(at)
+    }
+
+    fn memory_at(&self, at: u32) -> Option<MemoryType> {
+        self.cur.state.memory_at(at)
+    }
+
+    fn event_at(&self, at: u32) -> Option<&Self::FuncType> {
+        self.cur.state.event_at(at)
+    }
+
+    fn global_at(&self, at: u32) -> Option<GlobalType> {
+        self.cur.state.global_at(at)
+    }
+
+    fn func_type_at(&self, type_idx: u32) -> Option<&Self::FuncType> {
+        self.cur.state.func_type_at(type_idx)
+    }
+
+    fn type_of_function(&self, func_idx: u32) -> Option<&Self::FuncType> {
+        self.cur.state.type_of_function(func_idx)
+    }
+
+    fn element_type_at(&self, at: u32) -> Option<Type> {
+        self.cur.state.element_type_at(at)
+    }
+
+    fn element_count(&self) -> u32 {
+        self.cur.state.element_count()
+    }
+
+    fn data_count(&self) -> u32 {
+        self.cur.state.data_count()
+    }
+
+    fn is_function_referenced(&self, idx: u32) -> bool {
+        self.cur.state.is_function_referenced(idx)
+    }
+}
+
 fn combine_type_sizes(offset: usize, a: u32, b: u32) -> Result<u32> {
     match a.checked_add(b) {
         Some(sum) if sum < MAX_TYPE_SIZE => Ok(sum),
@@ -1627,23 +1680,20 @@ impl WasmFeatures {
     }
 }
 
-/// The implementation of [`WasmModuleResources`] used by [`Validator`].
-pub struct ValidatorResources(Arc<ModuleState>);
-
-impl WasmModuleResources for ValidatorResources {
+impl WasmModuleResources for ModuleState {
     type FuncType = crate::FuncType;
 
     fn table_at(&self, at: u32) -> Option<TableType> {
-        self.0.tables.get(at as usize).cloned()
+        self.tables.get(at as usize).cloned()
     }
 
     fn memory_at(&self, at: u32) -> Option<MemoryType> {
-        self.0.memories.get(at as usize).cloned()
+        self.memories.get(at as usize).cloned()
     }
 
     fn event_at(&self, at: u32) -> Option<&Self::FuncType> {
-        let types = self.0.all_types.as_ref().unwrap();
-        let i = *self.0.events.get(at as usize)?;
+        let types = self.all_types.as_ref().unwrap();
+        let i = *self.events.get(at as usize)?;
         match &types[i] {
             TypeDef::Func(f) => Some(f),
             _ => None,
@@ -1651,12 +1701,12 @@ impl WasmModuleResources for ValidatorResources {
     }
 
     fn global_at(&self, at: u32) -> Option<GlobalType> {
-        self.0.globals.get(at as usize).cloned()
+        self.globals.get(at as usize).cloned()
     }
 
     fn func_type_at(&self, at: u32) -> Option<&Self::FuncType> {
-        let types = self.0.all_types.as_ref().unwrap();
-        let i = *self.0.types.get(at as usize)?;
+        let types = self.all_types.as_ref().unwrap();
+        let i = *self.types.get(at as usize)?;
         match &types[i] {
             TypeDef::Func(f) => Some(f),
             _ => None,
@@ -1664,8 +1714,8 @@ impl WasmModuleResources for ValidatorResources {
     }
 
     fn type_of_function(&self, at: u32) -> Option<&Self::FuncType> {
-        let types = self.0.all_types.as_ref().unwrap();
-        let i = *self.0.func_types.get(at as usize)?;
+        let types = self.all_types.as_ref().unwrap();
+        let i = *self.func_types.get(at as usize)?;
         match &types[i] {
             TypeDef::Func(f) => Some(f),
             _ => None,
@@ -1673,19 +1723,66 @@ impl WasmModuleResources for ValidatorResources {
     }
 
     fn element_type_at(&self, at: u32) -> Option<Type> {
-        self.0.element_types.get(at as usize).cloned()
+        self.element_types.get(at as usize).cloned()
     }
 
     fn element_count(&self) -> u32 {
-        self.0.element_types.len() as u32
+        self.element_types.len() as u32
     }
 
     fn data_count(&self) -> u32 {
-        self.0.data_count.unwrap_or(0)
+        self.data_count.unwrap_or(0)
     }
 
     fn is_function_referenced(&self, idx: u32) -> bool {
-        self.0.function_references.contains(&idx)
+        self.function_references.contains(&idx)
+    }
+}
+
+/// The implementation of [`WasmModuleResources`] used by [`Validator`].
+pub struct ValidatorResources(Arc<ModuleState>);
+
+impl WasmModuleResources for ValidatorResources {
+    type FuncType = crate::FuncType;
+
+    fn table_at(&self, at: u32) -> Option<TableType> {
+        self.0.table_at(at)
+    }
+
+    fn memory_at(&self, at: u32) -> Option<MemoryType> {
+        self.0.memory_at(at)
+    }
+
+    fn event_at(&self, at: u32) -> Option<&Self::FuncType> {
+        self.0.event_at(at)
+    }
+
+    fn global_at(&self, at: u32) -> Option<GlobalType> {
+        self.0.global_at(at)
+    }
+
+    fn func_type_at(&self, at: u32) -> Option<&Self::FuncType> {
+        self.0.func_type_at(at)
+    }
+
+    fn type_of_function(&self, at: u32) -> Option<&Self::FuncType> {
+        self.0.type_of_function(at)
+    }
+
+    fn element_type_at(&self, at: u32) -> Option<Type> {
+        self.0.element_type_at(at)
+    }
+
+    fn element_count(&self) -> u32 {
+        self.0.element_count()
+    }
+
+    fn data_count(&self) -> u32 {
+        self.0.data_count()
+    }
+
+    fn is_function_referenced(&self, idx: u32) -> bool {
+        self.0.is_function_referenced(idx)
     }
 }
 
@@ -1772,7 +1869,8 @@ impl NameSet {
         types: &mut SnapshotList<TypeDef>,
         desc: &str,
     ) -> Result<Option<usize>> {
-        self.type_size = combine_type_sizes(offset, self.type_size, ty.size(types))?;
+        self.type_size =
+            combine_type_sizes(offset, self.type_size, ty.size(types).saturating_add(1))?;
         let name = match name {
             Some(name) => name,
             // If the `name` is not provided then this is a module-linking style
@@ -1849,7 +1947,7 @@ impl NameSet {
 
 impl EntityType {
     fn size(&self, list: &SnapshotList<TypeDef>) -> u32 {
-        let recursive_size = match self {
+        match self {
             // Note that this function computes the size of the *type*, not the
             // size of the value, so these "leaves" all count as 1
             EntityType::Global(_) | EntityType::Memory(_) | EntityType::Table(_) => 1,
@@ -1860,12 +1958,11 @@ impl EntityType {
             | EntityType::Module(i)
             | EntityType::Instance(i)
             | EntityType::Event(i) => match &list[*i] {
-                TypeDef::Func(f) => (f.params.len() + f.returns.len()) as u32,
-                TypeDef::Module(m) => m.type_size,
+                TypeDef::Func(f) => 1 + (f.params.len() + f.returns.len()) as u32,
+                TypeDef::Module(m) => m.imports_size + m.exports_size,
                 TypeDef::Instance(i) => i.type_size,
             },
-        };
-        recursive_size.saturating_add(1)
+        }
     }
 }
 
